-- Maintenance Madness - 2019. Created by Arcitos. License: See mod page.

local mm_util = require("util.util")
local const = require("config-runtime")

local rand = math.random
local mceil = math.ceil
local mfloor = math.floor

local forcesWithMROdisabled = const.forcesWithMROdisabled

local entityTypesWithMROenabled = const.entityTypesWithMROenabled

local entityMaintencanceDemandByName = const.entityMaintencanceDemandByName

local entityMaintenanceDemandByType = const.entityMaintenanceDemandByType

local textColors = const.textColors

----- MAIN FUNCTIONS      (-> Event handlers: line 700+)

-- set up aux table for maintenance demands per entity prototype class
local function initMaintenanceDemandsTable(entityName, timeFactor)
	local resultTable = {}
	-- the maintenance demands are defined in this config table:
	local source = entityMaintencanceDemandByName[entityName] or entityMaintenanceDemandByType[game.entity_prototypes[entityName].type] -- entity specification > type specification 
	for level, costFactor in pairs(const.costFactor) do
		resultTable[level] = {}
		for _, ingredient in pairs(source) do
			local itemName = ingredient.name or ingredient[1]
			local amount = mm_util.getIngredientAmount(ingredient)
			if ingredient.type ~= "fluid" then 	
				-- add the maintenance demands for each class to all entities
				resultTable[level][itemName] = {}
				resultTable[level][itemName].probability = ingredient.probability
				resultTable[level][itemName].min_amount = mceil(amount * costFactor * timeFactor / 2) -- min 50% maintenance demand
				resultTable[level][itemName].max_amount = mceil(amount * costFactor * timeFactor * 1,5) -- max 150% maintenance demand
			end
		end
	end	
	--log(serpent.block(resultTable, {maxlevel= 4}))
	return resultTable	
end

-- set up aux table for repair demands per entity prototype
local function initRepairDemandsTable(entityName, timeFactor)
	local resultTable = {}
	local ingredients = game.recipe_prototypes[entityName].ingredients
	for level, costFactor in pairs(const.costFactor) do
		resultTable[level] = {}
		for _, ingredient in pairs(ingredients) do
			local itemName = ingredient.name or ingredient[1]
			local amount = mm_util.getIngredientAmount(ingredient)
			if ingredient.type ~= "fluid" then 	
				-- repair cost is a fraction of the entity's construction cost
				resultTable[level][itemName] = {}
				resultTable[level][itemName].probability = math.min(amount * costFactor * timeFactor / 10, 0.5) -- ingredients in small amounts are often expensive, therefore the probability for using them as spare parts is lowered
				resultTable[level][itemName].min_amount = mceil(amount * costFactor * timeFactor / 25) -- min 4% of construction cost
				resultTable[level][itemName].max_amount = mceil(amount * costFactor * timeFactor / 8) -- max 12,5% of construction cost
			end
		end
		-- some repair packs are always required 
		local repairPackName = "repair-pack"
		local healthFactor = mceil(game.entity_prototypes[entityName].max_health * costFactor * timeFactor / 150) -- sturdier entities need more repair packs
		local itemTypes = mceil(mm_util.getLength(resultTable[level]) * costFactor * timeFactor) -- the more different items required for construction, the more repair packs needed
		resultTable[level][repairPackName] = {}
		resultTable[level][repairPackName].probability = 1
		resultTable[level][repairPackName].min_amount = mceil((itemTypes + healthFactor) / 3)   
		resultTable[level][repairPackName].max_amount = mceil((itemTypes + healthFactor) / 1.5)
	end	
	--log(serpent.block(resultTable, {maxlevel= 4}))
	return resultTable
end


-- start with building of some aux tables to avoid repeated calculations + allowing for dynamic value modification
local function initAuxTables()
	global.repairItems = {} -- WIP
	for _, item in pairs(game.item_prototypes) do
		if item.type == "repair-tool" then
			table.insert(global.repairItems, item.name)
		end
	end
	
	global.entitiesWithMROenabled = {}
	for _, entity in pairs(game.entity_prototypes) do
		-- add all entities of a prototype class, if this class is enabled in mod config
		if entityTypesWithMROenabled[entity.type] ~= nil then
			-- add an entity only if it has an recipe and an item
			if game.item_prototypes[entity.name] and game.recipe_prototypes[entity.name] then
				--log("Entity geladen: "..entity.name.." mit Typ "..entity.type)
				local dimensions = mm_util.getBoundingBoxDimensions(entity.selection_box)
				--log(serpent.block(entity.selection_box, {maxlevel= 4}))
				--log(serpent.block(dimensions, {maxlevel= 4}))
				--log(serpent.block(mm_util.rotateBoundingBox(entity.selection_box), {maxlevel= 4}))
				local square = true
				-- if an entity hat a non square base, we'll later need two chests for maintenance / repair / overhaul demands, as chests are not rotatable
				if dimensions.width ~= dimensions.height then
					square = false
				end
				-- some entity types have event cycles with increased length
				local timeFactor = 1
				if entity.type == "solar-panel" or entity.type == "accumulator" then
					timeFactor = const.solarCycleTimeMultiplier
				end

				local newElement = {}
				newElement.active = entityTypesWithMROenabled[entity.type]
				newElement.maintenance = initMaintenanceDemandsTable(entity.name, timeFactor)
				newElement.repair = initRepairDemandsTable(entity.name, timeFactor)
				newElement.scrap = {["mm-scrapped-"..entity.name] = 1}
				newElement.replacement = {[entity.name] = 1}
				newElement.flyingTextShiftY = entity.collision_box.left_top.y/2 -- this will place the flying text more on top of the entity
				newElement.square = square
				newElement.timeFactor = timeFactor
				global.entitiesWithMROenabled[entity.name] = newElement
				--log(serpent.block(global.entitiesWithMROenabled[entity.name], {maxlevel= 4}))
			end
		end
	end			
	--log(serpent.block(global.entitiesWithMROenabled, {maxlevel= 4}))
	if game.item_prototypes["SiSi-solar-cell"] then -- compatibility with Simple Silicon, WIP
		table.insert(entityMaintenanceDemandByType["solar-panel"], {name = "SiSi-solar-cell", amount = 1, probability = 0.4})
	end
end

-- get or init persistent vars
local function init_global()
    global = global or {}

	initAuxTables()

    global.eventScheduleMRO = global.eventScheduleMRO or {}

	global.monitoredEntities = global.monitoredEntities or {} -- indexed by unit number
end

-- calculate the demands for an maintenance or repair event for this entity
local function calculateDemands(entityName, source)
	--log(serpent.block(global.entitiesWithMROenabled[entityName].repair, {maxlevel= 4}))
	local resultTable = {}
	for item, itemData in pairs (source) do 
		if itemData.probability == 1 or rand() < itemData.probability then
			resultTable[item] = rand(itemData.min_amount, itemData.max_amount)		
		end
	end
	return resultTable
end

-- check, if the item demand for a given entity has been fulfilled
local function checkCompletedDemands(entityID)
	local percentage = 0.00
	local p = 0.00
	local deliveredItems = global.monitoredEntities[entityID].connectedChest.get_inventory(defines.inventory.chest).get_contents()
	local requestedItems = global.monitoredEntities[entityID].requestedItems
	local distinctItems = 0
	for _, amount in pairs(requestedItems) do
		-- count different items
		distinctItems = distinctItems + 1
	end
	-- calculate the percentage of the demand fulfillment
	for item, amount in pairs(requestedItems) do
		if deliveredItems[item] ~= nil then
			p = (deliveredItems[item] / amount) / distinctItems
		else
			p = 0
		end
		percentage = percentage + p	
	end
	return percentage
end

-- all items in the given chest will be counted as "consumption" in the production tab
local function postItemsAsConsumption(chest)
	local inventory = chest.get_inventory(defines.inventory.chest).get_contents()
	for name, amount in pairs(inventory) do
		chest.force.item_production_statistics.on_flow(name, -amount)
	end
end

-- set up the next event for a given entity in the event schedule 
local function scheduleNextEvent(eventTick, elementData)	
	if global.eventScheduleMRO[eventTick] == nil then
		-- insert first element for this tick
		global.eventScheduleMRO[eventTick] = {} 
		global.eventScheduleMRO[eventTick].elements = {elementData}
		--global.eventScheduleMRO[eventTick].elements[#global.eventScheduleMRO[eventTick].elements + 1] = elementData
	else
		-- add another element
		global.eventScheduleMRO[eventTick].elements[#global.eventScheduleMRO[eventTick].elements + 1] = elementData
	end
end

-- replace a surrogate entity (faulty version of an accumulator or a solar panel) with its original entity
local function swapOldMachine(entity)
	local machine = entity
	local targetReference = global.monitoredEntities[machine.unit_number]
	local newOperableEntity = machine.surface.create_entity{
		name = targetReference.surrogatedEntityName, 
		position = machine.position, 
		force = machine.force,
		create_build_effect_smoke = false}
	if newOperableEntity.valid then
		newOperableEntity.energy = machine.energy
		if machine.health > 0 then
			newOperableEntity.health = machine.health
		end
		if machine.to_be_deconstructed(machine.force) then newOperableEntity.order_deconstruction(machine.force) end
		local newtargetReference = targetReference
		newtargetReference.surrogatedEntityName = nil
		newtargetReference.entity = newOperableEntity
		newtargetReference.ageing = targetReference.ageing
		newtargetReference.status = targetReference.status
		global.monitoredEntities[newOperableEntity.unit_number] = newtargetReference
		targetReference = global.monitoredEntities[newOperableEntity.unit_number]
		global.monitoredEntities[machine.unit_number] = nil
		machine.destroy()
		machine = targetReference.entity 
	end
	return machine
end

-- add a new entity to the event schedule, starting its maintenance cycle
local function initLifeCycle(entity)
	
	if entity and entity.valid then
		local entityID = entity.unit_number
		local entityProperties = global.entitiesWithMROenabled[entity.name]
		local nextTick = rand(mfloor(const.MTBM/2), mfloor(const.MTBM*1,5))
		local eventTick = game.tick + mfloor(entityProperties.timeFactor * nextTick)
		
		scheduleNextEvent(eventTick, {id = entityID, age = 0})

		global.monitoredEntities[entityID] = {entity = entity, ageing = 0, status = "operating", producedItems = 0, bonusCyclesRemaining = const.freeCycles, connectedChest = nil, connectedProxy = nil, requestedItems = nil} -- nextCheck = eventTick?
		--game.print("Neues Objekt in Liste. Nächste Aktion in "..(mfloor(nextTick/60)).."s")
	end	
end

local function startMaintenance(entityID)
	--game.print("Beginne Wartung von Maschine...")
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		-- safety catch
		if targetReference.surrogatedEntityName then 
			machine = swapOldMachine(machine) 
		end
		local machineProperties = global.entitiesWithMROenabled[machine.name]
		targetReference.status = "maintenance"
		-- check for rotation:
		local chestType = "mm-chest-"
		if not machineProperties.square and (machine.direction == defines.direction.east or machine.direction == defines.direction.west) then
			chestType = "mm-chest-rotated-"
		end
		local targetChest = machine.surface.create_entity{
			name = chestType..machine.name, 
			position = machine.position, 
			force = machine.force, 
			create_build_effect_smoke = false}
		targetChest.operable = false
		targetReference.connectedChest = targetChest
		
		machine.teleport(machine.position) -- a hacky way to put the machine above the invisible chest (otherwise inserters are able to access the chest), while maintaining the same unit_number
	
		local requestedItems = calculateDemands(machine.name, machineProperties.maintenance[const.maintenanceLevel])
		machine.surface.create_entity{
			name = "mm-maintenance-request-proxy", 
			position = targetChest.position, 
			target = targetChest, 
			force = targetChest.force, 
			modules = requestedItems}
		targetReference.requestedItems = requestedItems
		
		local checkForCompletionTick = game.tick +  (rand(mfloor(const.MTTM/2), mfloor(const.MTTM*1,5)))
		-- this value is not multiplied with timeFactor!
		
		global.monitoredEntities[entityID] = targetReference
		scheduleNextEvent(checkForCompletionTick, {id = entityID, age = targetReference.ageing})
		--game.print("Wartung gestartet. Nächste Aktion in "..(math.floor((checkForCompletionTick - game.tick)/60)).."s")
	end
end

local function finishMaintenance(entityID)
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local machine = global.monitoredEntities[entityID].entity
	if machine and machine.valid then
		global.monitoredEntities[entityID].status = "operating"
		local machineProperties = global.entitiesWithMROenabled[machine.name]
		local targetChest = global.monitoredEntities[entityID].connectedChest
		local missedMaintenanceMalus = const.maxMissedMaintenanceMalus
		local maintenanceCompletion = 0
		if targetChest and targetChest.valid then
			maintenanceCompletion = checkCompletedDemands(entityID)
			missedMaintenanceMalus = missedMaintenanceMalus * (1 - maintenanceCompletion)
			postItemsAsConsumption(targetChest)
			targetChest.destroy()
			global.monitoredEntities[entityID].requestedItems = nil
			global.monitoredEntities[entityID].connectedChest = nil
		end
		-- increase ageing -> increased probability for failures
		local newAgeing = global.monitoredEntities[entityID].ageing + machineProperties.timeFactor * (const.baseAgeing + missedMaintenanceMalus)
		global.monitoredEntities[entityID].ageing = newAgeing
		
		if const.showText then
			local color = mm_util.calculateTextColor(textColors.maintenance.min, textColors.maintenance.delta, maintenanceCompletion)
			local textPos = {x = machine.position.x, y = machine.position.y + machineProperties.flyingTextShiftY}
			machine.surface.create_entity{name = "mm-flying-text", position = textPos, text = {"mm-flying-text-maintenance-finished", math.floor(maintenanceCompletion*100)}, color = color, force = machine.force}
		end
		
		local eventTick = game.tick + mfloor(machineProperties.timeFactor * (rand(mfloor(const.MTBM/2), mfloor(const.MTBM*1.5))))

		scheduleNextEvent(eventTick, {id = entityID, age = newAgeing})
			
	 -- todo: was fehlt noch?
		--game.print("Wartung beendet. Neues Alter der Maschine: "..newAgeing)
	end
end

local function causeFailure(entityID)
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		local damageFactor = const.baseDamageFactor + rand() * const.maxDamageFactor * (targetReference.ageing / const.maxAge)
		machine.damage(mfloor(machine.prototype.max_health * damageFactor), machine.force)
		if not machine.valid then
			return -- machine has been killed by damage
		end
		
		--- solar panels or accumulators wont stop working if set to active = false. They need to be replaced by "broken" versions of themselves
		if machine.type == "solar-panel" or machine.type == "accumulator" and (#machine.circuit_connected_entities.red + #machine.circuit_connected_entities.green) == 0 then
			--log(serpent.block(targetReference, {maxlevel= 4}))
			--log(entityID)
			local newFaultyEntity = machine.surface.create_entity{
				name = "mm-faulty-"..machine.name, 
				position = machine.position, 
				force = machine.force, 
				create_build_effect_smoke = false}
			newFaultyEntity.health = machine.health
			newFaultyEntity.energy = machine.energy
			if machine.to_be_deconstructed(machine.force) then newFaultyEntity.order_deconstruction(machine.force) end
			local newtargetReference = targetReference
			newtargetReference.surrogatedEntityName = machine.name
			newtargetReference.entity = newFaultyEntity
			newtargetReference.ageing = targetReference.ageing
			newtargetReference.status = targetReference.status
			global.monitoredEntities[newFaultyEntity.unit_number] = newtargetReference
			targetReference = global.monitoredEntities[newFaultyEntity.unit_number]
			global.monitoredEntities[entityID] = nil
			machine.destroy()
			machine = targetReference.entity 
			entityID = newFaultyEntity.unit_number
			--log(serpent.block(targetReference, {maxlevel= 4}))
			--log(entityID)
			-- TODO: Netzwerk-Verbindungen wiederherstellen > nicht möglich.
		end
		targetReference.status = "malfunction"
		local machineName = targetReference.surrogatedEntityName or machine.name -- get name of surrogated machine (in case of solar panel or accumulator) or otherwise real machine name
		local machineProperties = global.entitiesWithMROenabled[machineName]
		
		machine.active = false
		machine.operable = false
		machine.rotatable = false
		
		-- check for rotation:
		local chestType = "mm-chest-"
		if not machineProperties.square and (machine.direction == defines.direction.east or machine.direction == defines.direction.west) then
			chestType = "mm-chest-rotated-"
		end
		local targetChest = machine.surface.create_entity{
			name = chestType..machineName, 
			position = machine.position, 
			force = machine.force,
			create_build_effect_smoke = false}
		targetChest.operable = false
		targetReference.connectedChest = targetChest
		machine.teleport(machine.position) -- a hacky way to put the machine above the invisible chest (otherwise inserters are able to access the chest), while maintaining the same unit_number
	
		local requestedItems = calculateDemands(machineName, machineProperties.repair[const.maintenanceLevel])
		targetReference.connectedProxy = machine.surface.create_entity{name = "mm-repair-request-proxy", position = targetChest.position, target = targetChest, force = targetChest.force, modules = requestedItems}
		targetReference.requestedItems = requestedItems	
		
		
		-- add an icon that is hidden until the requester icon is removed by fulfilling the request
		if machine.type == "solar-panel" or machine.type == "accumulator" then
			if const.showMIPsignForSolar then
				rendering.draw_animation{
					animation = "repair-in-progress-icon",
					render_layer = "entity-info-icon",
					target = targetReference.connectedChest, 
					target_offset = machine.prototype.alert_icon_shift or {0, 0}, 
					surface = machine.surface,
					forces = {machine.force}}
			end
		else
			rendering.draw_animation{
				animation = "repair-in-progress-icon",
				render_layer = "entity-info-icon",
				target = targetReference.connectedChest, 
				target_offset = machine.prototype.alert_icon_shift or {0, 0}, 
				surface = machine.surface,
				forces = {machine.force}}
		end
		
		if const.showText then
			local textPos = {x = machine.position.x, y = machine.position.y + machineProperties.flyingTextShiftY}
			machine.surface.create_entity{name = "mm-flying-text", position = textPos, text = {"mm-flying-text-failure"}, color = textColors.failure, force = machine.force}
		end
		
		local checkForCompletionTick = game.tick + mfloor(machineProperties.timeFactor * (const.MTTR + (const.MTTR * const.repairTimeModifier * targetReference.ageing / const.maxAge)))
		
		global.monitoredEntities[entityID] = targetReference
		scheduleNextEvent(checkForCompletionTick, {id = entityID, age = targetReference.ageing})
		--game.print("Maschine hat Fehlfunktion. Nächste Aktion in "..(math.floor(const.MTTR/60)).."s")
	end
end

local function checkRepairState(entityID)
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		local eventTick = 0
		local machineName = targetReference.surrogatedEntityName or machine.name -- get name of surrogated machine (in case of solar panel or accumulator) or real machine name
		local machineProperties = global.entitiesWithMROenabled[machineName]
		if not (targetReference.connectedProxy and targetReference.connectedProxy.valid) then
			local targetChest = targetReference.connectedChest
			postItemsAsConsumption(targetChest)
			targetChest.destroy()
			targetReference.requestedItems = nil
			targetReference.connectedProxy = nil
			targetReference.connectedChest = nil
			--- solar panels and accumulators have been replaced by broken versions - these entities must be swapped back :
			if targetReference.surrogatedEntityName then 
				machine = swapOldMachine(machine) 
				entityID = machine.unit_number
			end
			--[[
				--machine.type == "solar-panel" or machine.type == "accumulator" and (#machine.circuit_connected_entities.red + #machine.circuit_connected_entities.green) == 0 then
				--log(serpent.block(targetReference, {maxlevel= 4}))
				--log(entityID)
				local newOperableEntity = machine.surface.create_entity{name = machineName, position = machine.position, force = machine.force}
				newOperableEntity.energy = machine.energy
				if machine.to_be_deconstructed(machine.force) then newOperableEntity.order_deconstruction(machine.force) end
				local newtargetReference = targetReference
				newtargetReference.surrogatedEntityName = nil
				newtargetReference.entity = newOperableEntity
				newtargetReference.ageing = targetReference.ageing
				newtargetReference.status = targetReference.status
				global.monitoredEntities[newOperableEntity.unit_number] = newtargetReference
				targetReference = global.monitoredEntities[newOperableEntity.unit_number]
				machine.destroy()
				machine = targetReference.entity 
				entityID = newOperableEntity.unit_number
				--log(serpent.block(targetReference, {maxlevel= 4}))
				--log(entityID)
			end]]
			targetReference.ageing = targetReference.ageing + const.repairAgeMalus
			targetReference.status = "operating"
			machine.active = true
			machine.operable = true
			machine.rotatable = true
			
			if const.showText then
				local textPos = {x = machine.position.x, y = machine.position.y + machineProperties.flyingTextShiftY}
				machine.surface.create_entity{name = "mm-flying-text", position = textPos, text = {"mm-flying-text-machine-fixed"}, color = textColors.repair, force = machine.force}
			end
			
			eventTick = game.tick + mfloor(machineProperties.timeFactor * rand(mfloor(const.MTBM/2), mfloor(const.MTBM*1.5)))
			
		else 
		-- If machineProperties.active wait some time, then check again if all needed parts have been delivered
			eventTick = game.tick + mfloor(machineProperties.timeFactor * const.tickDelayCheckRepairState)
			if const.showText then
				local textPos = {x = machine.position.x, y = machine.position.y + machineProperties.flyingTextShiftY}
				machine.surface.create_entity{name = "mm-flying-text", position = textPos, text = {"mm-flying-text-failure-ongoing"}, color = textColors.failure2, force = machine.force}
			end
		end 
			
		global.monitoredEntities[entityID] = targetReference
		scheduleNextEvent(eventTick, {id = entityID, age = targetReference.ageing})		
	end
end

local function requestReplacementOfOldMachine(entityID)
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	local machineProperties = global.entitiesWithMROenabled[machine.name]
	if machine and machine.valid and targetReference.status == "operating" then
		
		-- safety catch
		if targetReference.surrogatedEntityName then 
			machine = swapOldMachine(machine) 
		end
		-- check for rotation:
		local chestType = "mm-chest-"
		if not machineProperties.square and (machine.direction == defines.direction.east or machine.direction == defines.direction.west) then
			chestType = "mm-chest-rotated-"
		end
		local targetChest = machine.surface.create_entity{
			name = chestType..machine.name, 
			position = machine.position, 
			force = machine.force,
			create_build_effect_smoke = false}
		targetReference.connectedChest = targetChest
				
		machine.teleport(machine.position) -- a hacky way to put the machine above the invisible chest (otherwise inserters are able to access the chest), while maintaining the same unit_number
	
		local requestedItems = machineProperties.replacement
		targetReference.connectedProxy = machine.surface.create_entity{name = "mm-replacement-request-proxy", position = targetChest.position, target = targetChest, force = targetChest.force, modules = requestedItems}
		targetReference.requestedItems = requestedItems	
		targetReference.status = "awaiting-replacement"
		
		global.monitoredEntities[entityID] = targetReference
		-- No additional event tick needed, as there is already one scheduled
	end
end

local function checkForScrapReturn(entityID)
-- Depending on the age of the mined machine, the player / robot will either get back the machine item, or its scrap version
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return nil
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		if rand(0, const.maxAge) < targetReference.ageing then
			local name = targetReference.surrogatedEntityName or machine.name
			return global.entitiesWithMROenabled[name].scrap	
		end
	end
	return nil
end

local function checkReplacementSuccess(entityID)
-- This is to check, if the requestes replacement machine item has been delivered or not.
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		local targetChest = targetReference.connectedChest
		if not (targetReference.connectedProxy and targetReference.connectedProxy.valid) then 
		-- An invalid proxy equals a fulfilled proxy
			postItemsAsConsumption(targetChest)			
			targetReference.ageing = 0		
			local scrap = global.entitiesWithMROenabled[machine.name].scrap	
			local returnedItem = nil
			if scrap ~= nil then
				for item, amount in pairs(scrap) do
					returnedItem = machine.surface.create_entity{name = "item-on-ground", position = machine.position, stack = {name = item, count = amount}, force = machine.force}
					if returnedItem.valid then
						returnedItem.order_deconstruction(machine.force)
						machine.force.item_production_statistics.on_flow(item, amount)
					end
				end
			end
		else
			targetReference.connectedProxy.destroy()
		end
		-- In every case, remove the request proxy. It will be recreated again in a later event
		targetReference.status = "operating"		
		targetChest.destroy()
		
		targetReference.requestedItems = nil
		targetReference.connectedProxy = nil
		targetReference.connectedChest = nil
		
		global.monitoredEntities[entityID] = targetReference
		-- No additional event tick needed, as there is already one scheduled
	end
end
	
local function stopMaintenanceCycle(entityID)
-- This function stops the need of maintenance for the given entity
	if entityID == nil or global.monitoredEntities[entityID] == nil then
		return 
	end
	local targetReference = global.monitoredEntities[entityID]
	local machine = targetReference.entity
	if machine and machine.valid then
		local eventTick = 0
		local machineName = targetReference.surrogatedEntityName or machine.name -- get name of surrogated machine (in case of solar panel or accumulator) or real machine name
		local targetChest = targetReference.connectedChest
		-- Remove the chest, if present
		if targetChest ~= nil then
			postItemsAsConsumption(targetChest)
			targetChest.destroy()			
			targetReference.requestedItems = nil
			targetReference.connectedProxy = nil
			targetReference.connectedChest = nil
		end
		-- Swap machine, if this was a surrogate
		if targetReference.surrogatedEntityName then 
			machine = swapOldMachine(machine) 
			entityID = machine.unit_number
		end

		if targetReference.status == "malfunction" then
			-- Reactivate the machine, if it had a breakdown
			machine.active = true
			machine.operable = true
			machine.rotatable = true
			
			-- Tell the player, that this machine has been reset
			if const.showText then
				local textPos = {x = machine.position.x, y = machine.position.y + global.entitiesWithMROenabled[machineName].flyingTextShiftY}
				machine.surface.create_entity{name = "mm-flying-text", position = textPos, text = {"mm-flying-text-machine-fixed"}, color = textColors.repair, force = machine.force}
			end
		end
		-- Remove the entity from the monitoring list
		global.monitoredEntities[entityID] = nil
	end
	return machine
end
	
local function processRemovedEntity(entityID)
	if entityID ~= nil and global.monitoredEntities[entityID] ~= nil then
		local mEntity = global.monitoredEntities[entityID]
		if mEntity.connectedChest ~= nil then
			mEntity.connectedChest.destroy()
		end
		global.monitoredEntities[entityID] = nil
	end
end

local function startMaintenanceCycleForExistingEntities(names)
	for _, surface in pairs(game.surfaces) do
		for _, entity in pairs(surface.find_entities_filtered{name = names}) do
			if forcesWithMROdisabled[entity.force] == nil then
				initLifeCycle(entity)
			end
		end
	end
end

local function stopMaintenanceCycleForExistingEntities(names)
	for _, surface in pairs(game.surfaces) do
		for _, entity in pairs(surface.find_entities_filtered{name = names}) do
			if forcesWithMROdisabled[entity.force] == nil then
				local entityID = entity.unit_number
				stopMaintenanceCycle(entityID)
				global.monitoredEntities[entityID] = nil
			end
		end
	end
end

commands.add_command("maintenance-mod-reload", {"mm-cmd-reload-help"}, function(param) 
	init_global()
	local player = game.players[param.player_index]
	mm_util.notification({"mm-notification-cmd-issued", {"mm-prefix"}, {"mm-cmd-reload-completed"}})
end)

----- EVENT HANDLERS :

-- check mod changes
script.on_configuration_changed(
function(data)
    -- midgame installation
    if data.mod_changes ~= nil and data.mod_changes["MaintenanceMadness"] ~= nil and data.mod_changes["MaintenanceMadness"].old_version == nil then
        -- anounce installation
        mm_util.notification({"mm-notification-midgame-update", {"mm-prefix"}, data.mod_changes["MaintenanceMadness"].new_version})

    -- midgame update
    elseif data.mod_changes ~= nil and data.mod_changes["MaintenanceMadness"] ~= nil and data.mod_changes["MaintenanceMadness"].old_version ~= nil then
        local oldver = data.mod_changes["MaintenanceMadness"].old_version
        local newver = data.mod_changes["MaintenanceMadness"].new_version
        mm_util.notification({"mm-notification-new-version", {"mm-prefix"}, oldver, newver})
    
    -- if other mods changed, e.g. added new entities OR mod startup settings changed (possibly removing or adding entites)
    elseif data.mod_changes ~= nil or data.mod_startup_settings_changed then
        for index, force in pairs(game.forces) do
			force.reset_recipes()
			force.reset_technologies()
			force.reset_technology_effects()
        end
	end 
	init_global()
end)

script.on_init(
function()
	init_global()
end)

script.on_event({defines.events.on_built_entity, defines.events.on_robot_built_entity},
function(event)
	entity = event.created_entity
	if forcesWithMROdisabled[entity.force.name] ~= true and global.entitiesWithMROenabled[entity.name] ~= nil and global.entitiesWithMROenabled[entity.name].active then
		initLifeCycle(entity)
	end
end)

script.on_event({defines.events.on_robot_mined_entity, defines.events.on_player_mined_entity},
function(event)
	local eid = event.entity.unit_number
	if eid ~= nil and global.monitoredEntities[eid] ~= nil then
		local scrap = checkForScrapReturn(eid)
		if scrap ~= nil then
			event.buffer.clear()
			for item, amount in pairs(scrap) do
				event.buffer.insert({name = item, count = amount})
			end
		end
		processRemovedEntity(eid) -- remove this entity from the monitoring list
	end
end)

script.on_event(defines.events.on_entity_died,
function(event)
	local entity = event.entity
	local eid = entity.unit_number
	if eid ~= nil and global.monitoredEntities[eid] ~= nil then
		if global.monitoredEntities[eid].surrogatedEntityName ~= nil then
			-- replace the machine and then destroy the replacement: This way, the ghost of the surrogated entity will be created 
			entity = stopMaintenanceCycle(eid)
			
			if event.cause ~= nil then
				entity.die(event.force, event.cause)
			else
				entity.die(event.force)
			end
		else
			stopMaintenanceCycle(eid)
		end
	end
end)

script.on_event({defines.events.on_player_setup_blueprint, defines.events.on_player_configured_blueprint},
function(event)
	-- If a player selects faulty versions of solar panels or accumulators for a blueprint, they need to be replaced by normal versions of themselves
	local player = game.players[event.player_index]
	if not player.valid then 
		return 
	end
	local bp = player.blueprint_to_setup
		if not bp.valid or not bp.valid_for_read then
			bp = player.cursor_stack
			if not bp.valid or not bp.valid_for_read then
				return
			end
	end
	if bp.name ~= "blueprint" then 
		return 
	end	
	local entities = bp.get_blueprint_entities()
	if not entities then 
		return 
	end
	local modified = false
	for _, entity in pairs (entities) do		
		if string.sub(entity.name, 1, 10) == "mm-faulty-" then 
			entity.name = string.gsub(entity.name, "mm%-faulty%-", "")
			
			modified = true
		end
	end
	if modified then
		if bp.blueprint_icons then
		-- Also swap the blueprint icons
			local icons = bp.blueprint_icons
			for _, icon in pairs (icons) do		
				if string.sub(icon.signal.name, 1, 10) == "mm-faulty-" then 
					icon.signal.name = string.gsub(icon.signal.name, "mm%-faulty%-", "")
				end
			end
			bp.blueprint_icons = icons
		end
		bp.set_blueprint_entities(entities)
	
	end
		
end)

script.on_event(defines.events.on_player_pipette,
function(event)
	local player = game.players[event.player_index]
	local item = event.item
	-- if the player pipettes faulty entities (e.g. solar panels), the items placed in the player's cursor need to be exchanged
	if player and player.valid and item and string.sub(item.name, 1, 10) == "mm-faulty-" then 
		item = game.item_prototypes[string.gsub(item.name, "mm%-faulty%-", "")]
		local inventory = player.get_main_inventory().get_contents()
		local item_amount = inventory[item.name] or 0
		if item_amount > 0 and player.cursor_stack.can_set_stack then
			-- Remove stack from inventory and put it into cursor
			player.cursor_stack.set_stack({name = item.name, count = item_amount})
			player.get_main_inventory().remove({name = item.name, count = item_amount})
		else
			if event.used_cheat_mode and player.cursor_stack.can_set_stack then 
				-- In cheat mode, create a new stack
				player.cursor_stack.set_stack({name = item.name, count = item.stack_size})
			else
				-- Ghost cursor
				player.cursor_ghost = item
			end
		end
	end
end)

script.on_event(defines.events.on_runtime_mod_setting_changed,
function(event)
	local s = event.setting
	if event.setting_type == "runtime-global" then
		local initMaintenancePrototypes = {}
		local stopMaintenancePrototypes = {}
		if s == "mm-mod-active" then
		-- This 
			const.maintenanceActivated = settings.global["mm-mod-active"].value
			if const.maintenanceActivated then
			-- Initiate maintenance for all enabled prototypes
				initMaintenancePrototypes["assembling-machine"] = true 
				initMaintenancePrototypes["furnace"] = true 
				if settings.global["mm-include-accumulators"].value then 
					initMaintenancePrototypes["accumulator"] = true end
				if settings.global["mm-include-solar-panels"].value then 
					initMaintenancePrototypes["solar-panel"] = true end
				if settings.global["mm-include-reactors"].value then 
					initMaintenancePrototypes["reactor"] = true end
				if settings.global["mm-include-boiler-and-generators"].value then 
					initMaintenancePrototypes["boiler"] = true 
					initMaintenancePrototypes["generator"] = true end
				if settings.global["mm-include-miners"].value then 
					initMaintenancePrototypes["mining-drill"] = true end
				if settings.global["mm-include-roboports"].value then 
					initMaintenancePrototypes["roboport"] = true end
				if settings.global["mm-include-beacons"].value then 
					initMaintenancePrototypes["beacon"] = true end
				if settings.global["mm-include-labs"].value then 
					initMaintenancePrototypes["lab"] = true end
			else
			-- stop maintenance for all prototypes
				stopMaintenancePrototypes["assembling-machine"] = true 
				stopMaintenancePrototypes["furnace"] = true 
				stopMaintenancePrototypes["accumulator"] = true 
				stopMaintenancePrototypes["solar-panel"] = true
				stopMaintenancePrototypes["reactor"] = true
				stopMaintenancePrototypes["boiler"]  = true 
				stopMaintenancePrototypes["generator"] = true
				stopMaintenancePrototypes["mining-drill"] = true
				stopMaintenancePrototypes["roboport"] = true
				stopMaintenancePrototypes["beacon"] = true
				stopMaintenancePrototypes["lab"] = true
			end			
		elseif s == "mm-maintenance-cost" then 
			const.maintenanceLevel = settings.global["mm-maintenance-cost"].value
		elseif s == "mm-show-flying-text" then
			const.showText = settings.global["mm-show-flying-text"].value
		elseif s == "mm-show-maintenance-in-progress-sign-for-solar" then
			const.showMIPsignForSolar = settings.global["mm-show-maintenance-in-progress-sign-for-solar"].value			
		elseif s == "mm-mean-time-between-maintenance-events" then
			const.MTBM = mfloor(3600 * settings.global["mm-mean-time-between-maintenance-events"].value) -- 3600 ticks = 1 min
		elseif s == "mm-mean-time-to-maintain" then
			const.MTTM = mfloor(3600 * settings.global["mm-mean-time-to-maintain"].value)
		elseif s == "mm-mean-time-to-repair" then	
			const.MTTM = mfloor(3600 * settings.global["mm-mean-time-to-repair"].value)			
		elseif s == "mm-include-accumulators" then
			if settings.global["mm-include-accumulators"].value then initMaintenancePrototypes["accumulator"] = true 
			else stopMaintenancePrototypes["accumulator"] = true end
		elseif s == "mm-include-solar-panels" then
			if settings.global["mm-include-solar-panels"].value then initMaintenancePrototypes["solar-panel"] = true 
			else stopMaintenancePrototypes["solar-panel"] = true end
		elseif s == "mm-include-reactors" then
			if settings.global["mm-include-reactors"].value then initMaintenancePrototypes["reactor"] = true  
			else stopMaintenancePrototypes["reactor"] = true end
		elseif s == "mm-include-boiler-and-generators" then
			if settings.global["mm-include-boiler-and-generators"].value then initMaintenancePrototypes["boiler"]  = true
			else stopMaintenancePrototypes["boiler"] = true end
			if settings.global["mm-include-boiler-and-generators"].value then initMaintenancePrototypes["generator"] = true 
			else stopMaintenancePrototypes["generator"] = true end
		elseif s == "mm-include-miners" then
			if settings.global["mm-include-miners"].value then initMaintenancePrototypes["mining-drill"] = true  
			else stopMaintenancePrototypes["mining-drill"] = true end
		elseif s == "mm-include-roboports" then
			if settings.global["mm-include-roboports"].value then initMaintenancePrototypes["roboport"] = true
			else stopMaintenancePrototypes["roboport"] = true end
		elseif s == "mm-include-beacons" then
			if settings.global["mm-include-beacons"].value then initMaintenancePrototypes["beacon"] = true
			else stopMaintenancePrototypes["beacon"] = true end		
		elseif s == "mm-include-labs" then
			if settings.global["mm-include-labs"].value then initMaintenancePrototypes["lab"] = true
			else stopMaintenancePrototypes["lab"] = true end		
		elseif s == "mm-solar-factor" then
			const.solarCycleTimeMultiplier = settings.global["mm-solar-factor"].value
		elseif s == "mm-max-age" then
			const.maxAge = settings.global["mm-max-age"].value
		elseif s == "mm-replacement-age" then
			const.replacementAge = settings.global["mm-replacement-age"].value
		elseif s == "mm-base-ageing" then
			const.baseAgeing = settings.global["mm-base-ageing"].value
		elseif s == "mm-missed-maintenance-malus" then
			const.maxMissedMaintenanceMalus = settings.global["mm-missed-maintenance-malus"].value
			const.repairAgeMalus = const.maxMissedMaintenanceMalus * 0,5
		end
		--log(serpent.block(initMaintenancePrototypes, {maxlevel= 4}))
		--log(serpent.block(stopMaintenancePrototypes, {maxlevel= 4}))
		if const.maintenanceActivated and mm_util.getLength(initMaintenancePrototypes) > 0 then
			local initEntity = {}
			for _, entity in pairs(game.entity_prototypes) do
				--log("Liste Entity: "..entity.name.." // "..entity.type)
				if initMaintenancePrototypes[entity.type] then
					if game.item_prototypes[entity.name] and game.recipe_prototypes[entity.name] then
						global.entitiesWithMROenabled[entity.name].active = true
						initEntity[#initEntity+1] = entity.name
					end
				end
			end
			if initEntity ~= nil then 
				--log(serpent.block(initEntity, {maxlevel= 4}))
				startMaintenanceCycleForExistingEntities(initEntity)  -- search for entities on map that now need to be maintained
				--log(serpent.block(global.entitiesWithMROenabled, {maxlevel= 4}))
			end
		end
		if  mm_util.getLength(stopMaintenancePrototypes) > 0 then
			if not const.maintenanceActivated then -- mod disabled, initiate instant stop. This may lag the game
				local stopEntities = {}
				for _, entity in pairs(game.entity_prototypes) do
					--log("Liste Entity: "..entity.name.." // "..entity.type)
					if stopMaintenancePrototypes[entity.type] then
						if game.item_prototypes[entity.name] and game.recipe_prototypes[entity.name] then
							global.entitiesWithMROenabled[entity.name].active = false
							stopEntities[#stopEntities+1] = entity.name
						-- this covers faulty versions of accus or solar panels
						elseif string.sub(entity.name, 1, 10) == "mm-faulty-" then
							stopEntities[#stopEntities+1] = entity.name
						end
					end
				end
				if stopEntities ~= nil then 
					--log(serpent.block(stopEntities, {maxlevel= 4}))
					stopMaintenanceCycleForExistingEntities(stopEntities)  -- search for entities on map that now need to be maintained
					--log(serpent.block(global.entitiesWithMROenabled, {maxlevel= 4}))
				end
				-- RESET EVENT SCHEDULE
				global.eventScheduleMRO = {}
			else -- slow fade out of existing maintenance requests for all entities with maintenance disabled 
				for entityName, entityData in pairs(global.entitiesWithMROenabled) do
					if stopMaintenancePrototypes[game.entity_prototypes[entityName].type] then
						entityData.active = false
					end
				end
			end
		end
	end		
end)

script.on_event(defines.events.on_tick,
function(event)
	--[[
	if event.tick%(60*60) == 0 then
		log("---")
		log("DEBUG: #global.monitoredEntities ="..mm_util.getLength(global.monitoredEntities))
		log("DEBUG: #global.eventScheduleMRO ="..mm_util.getLength(global.eventScheduleMRO))
	end
	]]
	if global.eventScheduleMRO and global.eventScheduleMRO[event.tick] ~= nil then
		for _, element in ipairs(global.eventScheduleMRO[event.tick].elements) do
			--game.print(_.." ,"..element.id.." ,"..element.age)
			local entityData = global.monitoredEntities[element.id]
			--log(serpent.block(element, {maxlevel= 2}))
			--log(serpent.block(entityData, {maxlevel= 3}))
			if entityData ~= nil and entityData.entity and entityData.entity.valid then
				local entityName = entityData.surrogatedEntityName or entityData.entity.name
				if not global.entitiesWithMROenabled[entityName].active then
					-- Stop maintenance for this entity
					stopMaintenanceCycle(element.id)
					goto continue_with_next_element
				end
				-- Handle bonus cycles. This has two purposes: First, the player has to deal with maintenance only after a certain amount of time (this is circa one hour, assuming default configuration)
				-- Second, the length of one bonus cycle is also randomised. This dampens the overall demand for maintenance materials and avoids huge initial demand spikes which would easily overwhelm your logistics capabilities
				if entityData.bonusCyclesRemaining ~= nil and entityData.bonusCyclesRemaining > 0 then
					entityData.bonusCyclesRemaining = entityData.bonusCyclesRemaining - 1
					local entityProperties = global.entitiesWithMROenabled[entityName]
					local nextTick = rand(mfloor(const.MTBM/2), mfloor(const.MTBM*1,5))
					local eventTick = game.tick + mfloor(entityProperties.timeFactor * nextTick)
					local newAgeing =  global.monitoredEntities[element.id].ageing + entityProperties.timeFactor * const.baseAgeing
		
					scheduleNextEvent(eventTick, {id = element.id, age = newAgeing})
					goto continue_with_next_element
				end
				--game.print("->")
				local entityAge = element.age
				-- if entityData.producedItems == entityData >>> TODO?
				if entityData.status == "awaiting-replacement" then
					checkReplacementSuccess(element.id)
					entityAge = global.monitoredEntities[element.id].ageing
				end
				if entityData.status == "maintenance" then
					--game.print("Objekt-ID: "..element.id.." / Alter: "..element.age.." / Status: "..entityData.status)
					finishMaintenance(element.id)
				elseif entityData.status == "operating" then
					if rand(0, const.maxAge) < entityAge * const.repairProbabilityModifier then 
						--game.print("Objekt-ID: "..element.id.." / Alter: "..element.age.." / Status: "..entityData.status)
						causeFailure(element.id)
					else
						--game.print("Objekt-ID: "..element.id.." / Alter: "..element.age.." / Status: "..entityData.status)
						startMaintenance(element.id)
					end
				elseif entityData.status == "malfunction" then
					checkRepairState(element.id)					
				end
				if entityData.status == "operating" and entityAge >= const.replacementAge then 
					requestReplacementOfOldMachine(element.id)
				end
			else
				-- If entity no longer valid OR maintenance for entity disabled - ignore this event
				global.monitoredEntities[element.id] = nil
			end
			::continue_with_next_element::
		end
		global.eventScheduleMRO[event.tick] = nil
	end				
end)